% This script demonstrates the use of the code to solve maximum
% likelihood estimation of the parameters ell and alpha in the power
% law generalized covarinace kernel for the following configuration of
% the points:
%
%   2D, regular grid, containing missing regions in disc shape.
%
% This script requires defining an auxiliary function
%
%   ex10_aux_func
%
% to appropriately handle matrix-vector multiplications and a wrapper
%
%   ex10_score_func
%
% to incorporate these multiplications.
%
% Note that due to the appearance of the likelihood surface the
% nonlinear solver fsolve does not work in this case. We reformulate
% the nonlinear equations in ell and alpha into a nonlinear equation
% in alpha only, and use bracketing to solve for alpha.
%

%Author: J. Chen
%$Date: 2012/01/19 18:09:43$

clear all
addpath(genpath('../src'))


%---------------- PARAMETERS ----------------
dim = [32;32];         % grid: size
T = [100;100];         % grid: physical dimension
discs = {[40;60;20]};  % grid: discs ([x-coord, y-coord, radius])

alpha = 1.5;           % kernel parameter
ell = [7;10];          % kernel parameter

alpha0 = [1.3;1.7];    % nonlinear solver: initial guess (bracket)
ell0 = [10;7];         % nonlinear solver: initial guess
N = 50;                % nonlinear solver: # samples in SAA
tol_fsolve = 1e-4;     % nonlinear solver: tol
tol_fzero = 1e-8;      % nonlinear solver: tol

m = 0;                 % linear solver: deflation dimension
maxit = 200;           % linear solver: maxit
tol_cg = 1e-9;         % linear solver: tol

N2 = 100;              % Fisher trace approx: # samples in SAA

num_filter = round((alpha+length(dim))/4);  % filtering

RandStream.setDefaultStream(RandStream('mt19937ar','seed',1420));


%---------------- GENERATE GP SAMPLE ----------------
fprintf(1, '--> Generating GP sample ...\n');

% generate covariance matrix for regular grid based on current ell
tK = gen_K_power_regular(dim, T, ell, alpha);

% filter K using Laplace filter
% this has to be done to ensure positive definiteness of K
tK = lapFilterMat2d(tK, num_filter);

% sample GP using the embedding method
[y, inc] = sampl_embed(tK, 4);

% find points corresponding to the missing regions
invalid_pts = find_invalid_pts(dim, T, discs, num_filter);

% take care of y
y(invalid_pts) = [];


%---------------- GENERATE RADEMACHER VECTORS ----------------
fprintf(1, '--> Generating Rademacher vectors ...\n');

% generate N independent Rademacher vectors
n = length(y);
u = sign(randn(n, N));
u(u==0) = 1;


%---------------- SOLVE MAX LOG-LIKELIHOOD ----------------
fprintf(1, '--> Solving max likelihood ...\n');

% solve the SAA of the score equation w.r.t. alpha
% do not use deflation in inner linear solver (m = 0)
tic;

% find alpha using fzero.
[alpha_est] = ...
    fzero(@(alpha)ex10_score_alpha(dim, T, ell0, alpha, num_filter, ...
                                   invalid_pts, y, u, m, maxit, tol_cg, ...
                                   tol_fsolve), ...
           alpha0, optimset('Display','iter','TolX',tol_fzero));

% Optimal alpha found. Compute the corresponding ell values.
[ell_est] = ...
    fsolve(@(ell)ex10_score_func(dim, T, ell, alpha_est, num_filter, ...
                                 invalid_pts, y, u, m, maxit, tol_cg,'ell'), ...
           ell0, optimset('Display','off','TolX',tol_fsolve));

elapsed_time = toc;


%---------------- COMPUTE COVARIANCE OF MLE ----------------
fprintf(1, '--> Computing covariance of MLE ...\n');

% compute covariance of MLE
[K, dK] = ex10_aux_func(dim, T, ell, alpha, num_filter, invalid_pts, 'mat');
[I0, V0] = score_func_mle_covar(K, dK, N);

% compute covariance of MLE (approximately)
u2 = sign(randn(n, N2));
u2(u2==0) = 1;

[Kfun, dKfun, Mfun] = ex10_aux_func(dim, T, ell, alpha, ...
                                    num_filter, invalid_pts, 'func');
[I, V] = score_func_mle_covar_approx(Kfun, dKfun, Mfun, N, ...
                                     u2, m, maxit, tol_cg);


%---------------- PRINT RESULTS ----------------
fprintf(1, '--> Done.\n\n');
fprintf(1, 'MAX LIKELIHOOD RESULTS:\n');
fprintf(1, 'Truth:      ell = %8.5f, %8.5f; alpha = %8.5f\n', ell, alpha);
fprintf(1, 'Init guess: ell = %8.5f, %8.5f; alpha in [%7.5f, %7.5f]\n', ell0, alpha0);
fprintf(1, 'Estimate:   ell = %8.5f, %8.5f; alpha = %8.5f\n', ell_est, alpha_est);
fprintf(1, 'Walltime:   %g sec\n', elapsed_time);
%{
fprintf(1, 'MLE covariance (inverse of Fisher matrix I):\n\n');
invI0 = inv(I0);
disp(invI0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI0));

fprintf(1, 'MLE covariance with SAA (inv(A)*B*inv(A)):\n\n');
disp(V0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V0));
%}
fprintf(1, 'MLE covariance (approximately):\n\n');
invI = inv(I);
disp(invI);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI));

fprintf(1, 'MLE covariance with SAA (approximately):\n\n');
disp(V);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V));
