% This script demonstrates the use of the code to solve maximum
% likelihood estimation of the scale parameter ell in the Matern
% covariance kernel for the following configuration of the points:
%
%   3D, regular grid.
%
% This script requires defining an auxiliary function
%
%   ex3_aux_func
%
% to appropriately handle matrix-vector multiplications and a wrapper
%
%   ex3_score_func
%
% to incorporate these multiplications.
%

%Author: J. Chen
%$Date: 2012/01/19 18:10:50$

clear all
addpath(genpath('../src'))


%---------------- PARAMETERS ----------------
dim = [10;10;10];   % grid: size
T = [100;100;100];  % grid: physical dimension

nu = 1;             % Matern: smoothness
ell = [7;10;13];    % Matern: scale

ell0 = [10;13;7];   % nonlinear solver: initial guess
N = 50;             % nonlinear solver: # samples in SAA
tol_fsolve = 1e-4;  % nonlinear solver: tol

m = 0;              % linear solver: deflation dimension
maxit = 200;        % linear solver: maxit
tol_cg = 1e-9;      % linear solver: tol

N2 = 100;           % Fisher trace approx: # samples in SAA

num_filter = round((nu*2+length(dim))/4);  % filtering

RandStream.setDefaultStream(RandStream('mt19937ar','seed',1420));


%---------------- GENERATE GP SAMPLE ----------------
fprintf(1, '--> Generating GP sample ...\n');

% generate covariance matrix for regular grid based on true ell
tK = gen_K_matern_regular(dim, T, ell, nu);

% sample GP using the embedding method
[y, inc] = sampl_embed(tK, 4);

% filter the sample using Laplace filter
y = lapFilterVec3d(reshape(y,dim'), num_filter);
y = y(:);


%---------------- GENERATE RADEMACHER VECTORS ----------------
fprintf(1, '--> Generating Rademacher vectors ...\n');

% generate N independent Rademacher vectors
u = sign(randn(prod(dim-2*num_filter), N));
u(u==0) = 1;


%---------------- SOLVE MAX LOG-LIKELIHOOD ----------------
fprintf(1, '--> Solving max likelihood ...\n');

% solve the SAA of the score equation
% do not use deflation in inner linear solver (m = 0)
tic;
[ell_est, fval, exitflag, output, jacob] = ...
    fsolve(@(ell)ex3_score_func(dim, T, ell, nu, num_filter, ...
                                y, u, m, maxit, tol_cg), ...
           ell0, optimset('Display','iter','TolX',tol_fsolve));
elapsed_time = toc;


%---------------- COMPUTE COVARIANCE OF SAA ----------------
fprintf(1, '--> Computing covariance of SAA ...\n');

% compute covariance of SAA
[Kfun, dKfun, Mfun] = ex3_aux_func(dim, T, ell_est, nu, num_filter, 'func');
Gamma = score_func_saa_covar(Kfun, dKfun, Mfun, ...
                             y, u, m, maxit, tol_cg, jacob);


%---------------- COMPUTE COVARIANCE OF MLE ----------------
fprintf(1, '--> Computing covariance of MLE ...\n');
%{
% compute covariance of MLE
[K, dK] = ex3_aux_func(dim, T, ell, nu, num_filter, 'mat');
[I0, V0] = score_func_mle_covar(K, dK, N);
%}
% compute covariance of MLE (approximately)
u2 = sign(randn(prod(dim-2*num_filter), N2));
u2(u2==0) = 1;

[Kfun, dKfun, Mfun] = ex3_aux_func(dim, T, ell, nu, num_filter, 'func');
[I, V] = score_func_mle_covar_approx(Kfun, dKfun, Mfun, N, ...
                                     u2, m, maxit, tol_cg);


%---------------- PRINT RESULTS ----------------
fprintf(1, '--> Done.\n\n');
fprintf(1, 'MAX LIKELIHOOD RESULTS:\n');
fprintf(1, 'Truth:      ell = %8.5f, %8.5f, %8.5f\n', ell);
fprintf(1, 'Init guess: ell = %8.5f, %8.5f, %8.5f\n', ell0);
fprintf(1, 'Estimate:   ell = %8.5f, %8.5f, %8.5f\n', ell_est);
fprintf(1, 'Walltime:   %g sec\n', elapsed_time);
%{
fprintf(1, 'MLE covariance (inverse of Fisher matrix I):\n\n');
invI0 = inv(I0);
disp(invI0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI0));

fprintf(1, 'MLE covariance with SAA (inv(A)*B*inv(A)):\n\n');
disp(V0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V0));
%}
fprintf(1, 'MLE variance (approximately):\n\n');
invI = inv(I);
disp(invI);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI));

fprintf(1, 'MLE variance with SAA (approximately):\n\n');
disp(V);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V));

fprintf(1, 'SAA covariance (inv(jacob)''*V*inv(jacob)/N):\n\n');
disp(Gamma);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(Gamma));
