function [coord, tri] = gen_coord_deformed_2d(dim, T)
%function [coord, tri] = gen_coord_deformed_2d(dim, T)
%
% This function generates a set of points based on a deformed regular
% grid. For the shape and the description of the grid, see [1, Fig.
% 6.3]. A mesh of the points is also returned.
%
% INPUTS
%   dim:   Grid size. E.g., [100;100] denotes a 100x100 grid.
%   T:     Physical domain. E.g., [50;50].
%
% OUTPUTS
%   coord: Coordinates of the points, including boundary.
%   tri:   Triangles of the mesh. Each row represents one triangle.
%
% [1] Jie Chen, Mihai Anitescu and Yousef Saad. Computing $f(A)b$ via
%     Least Squares Polynomial Approximations. SIAM Journal on
%     Scientific Computing, 33(1):195--222, 2011.
%

%Author: J. Chen
%$Date: 2011/11/23 11:44:31$

x = (-1:dim(2))/(dim(2)-1) - 0.5;
y = (-1:dim(1))/(dim(1)-1) - 0.5;
[X,Y] = meshgrid(x,y);

% scale the y coordinate of the grid points
suppx = max(max(X));
inffx = min(min(X));
mullx = 2 * (X-ones(size(X))*(suppx+inffx)/2) / (suppx-inffx);
mullx = ones(size(X)) - mullx.*mullx/2;
Y = mullx.*Y;

if length(T) == 1
  T = repmat(T,1,2);
end

X = X(:) * T(1);
Y = Y(:) * T(2);

coord = [X Y];
if nargout > 1
  tri = delaunay(X, Y);
end
