function [alpha, beta, gamma, res, iter] = ls_poly(f, t, tt, k, tol)
%function [alpha, beta, gamma, res, iter] = ls_poly(f, t, tt, k, tol)
%
% This function builds the least squares polynomial approximation to f
% by using a cubic spline. This function is used by sampl_sqrt.
%
% INPUTS
%   f:    Function handle. E.g., @sqrt.
%   t:    Knots. [t(1) t(end)] is the interval for approximation.
%   tt:   Locations to check approximation error.
%   k:    Maximum degree of approximating polynomial.
%   tol:  Tolerance in error at locations tt.
%
% OUTPUTS
%   alpha, beta, gamma: Coefficients in the Stieltjes procedure.
%   res:  Error at locations tt.
%   iter: Degree of approximating polynomial.
%   

%Author: J. Chen
%$Date: 2011/11/23 11:49:04$

m = length(tt);
b = ones(m,1);
n = length(t)-1;

xi    = zeros(n,4);
sigma = zeros(n,k+1);
eta   = zeros(n,k+1);
mu    = zeros(n,k+2);
mu1   = zeros(n,k+2);

h = reshape(diff(t)/2, n,1);
g = reshape((t(1:end-1)+t(2:end))/2, n,1);

pp = spline(t, f(t));
d = pp.coefs(:,1);
c = pp.coefs(:,2);
e = pp.coefs(:,3);
a = pp.coefs(:,4);
dh3 = d.*h.^3;
ch2 = c.*h.^2;
eh  = e.*h;
xi(:,1) =  5/2*dh3 + 3/2*ch2 + eh + a;
xi(:,2) = 15/4*dh3 +   2*ch2 + eh;
xi(:,3) =  3/2*dh3 + 1/2*ch2;
xi(:,4) =  1/4*dh3;

alpha = zeros(k,1);
beta = zeros(k+1,1);
gamma = zeros(k+1,1);

beta(1) = sqrt(n*pi);
mu(:,1) = 1/beta(1);
gamma(1) = pi * sum(xi(:,1).*mu(:,1));
v = b/beta(1);
z = gamma(1) * v;

for j = 1:k
  sigma(:,1) = h/2.*mu(:,2) + g.*mu(:,1);
  for p = 1:j
    sigma(:,p+1) = h/2.*(mu(:,p)+mu(:,p+2)) + g.*mu(:,p+1);
  end
  
  alpha(j) = sum( sigma(:,1).*mu(:,1) + h/4.*mu(:,1).*mu(:,2) );
  for p = 1:j
    alpha(j) = alpha(j) + (sigma(:,p+1)'*mu(:,p+1))/2;
  end
  alpha(j) = pi * alpha(j);

  if j ~= 1
    for p = 0:j
      eta(:,p+1) = sigma(:,p+1)-alpha(j)*mu(:,p+1)-beta(j)*mu0(:,p+1);
    end
  else
    for p = 0:j
      eta(:,p+1) = sigma(:,p+1)-alpha(j)*mu(:,p+1);
    end
  end
  
  beta(j+1) = sum(eta(:,1).^2) + sum((eta(:,2)+h/2.*mu(:,1)).^2)/2;
  for p = 2:j
    beta(j+1) = beta(j+1) + (eta(:,p+1)'*eta(:,p+1))/2;
  end
  beta(j+1) = sqrt(pi*beta(j+1));
  
  mu1(:,2) = (eta(:,2) + h/2.*mu(:,1)) / beta(j+1);
  for p = [0 2:j]
    mu1(:,p+1) = eta(:,p+1) / beta(j+1);
  end
  
  if j ~= 1
    v1 = (tt.*v - alpha(j)*v - beta(j)*v0) / beta(j+1);
  else
    v1 = (tt.*v - alpha(j)*v) / beta(j+1);
  end

  gamma(j+1) = sum( xi(:,1).*mu1(:,1) );
  for p = 1:min(j,3)
    gamma(j+1) = gamma(j+1) + sum(xi(:,p+1).*mu1(:,p+1))/2;
  end
  gamma(j+1) = pi * gamma(j+1);
  
  z = z + gamma(j+1)*v1;
  
  % check residual
  iter = j;
  res(j) = max(abs( f(tt)-z ));
  if res(j) < tol
    alpha = alpha(1:iter);
    beta = beta(1:iter+1);
    gamma = gamma(1:iter+1);
    res = res(1:iter);
    return;
  end

  % update
  v0 = v;
  v = v1;
  mu0 = mu;
  mu = mu1;
end

if iter == k
  fprintf(1, '  LS_POLY: max degree reached. ');
else
  fprintf(1, '  LS_POLY: successful. ');
end
fprintf(1, 'poly degree = %d, res = %g.\n', iter, res(end));
