function [y] = sampl_sqrt(Kmult, Lmult, LLTinv, n, degree, tol)
%function [y] = sampl_sqrt(Kmult, Lmult, LLTinv, n, degree, tol)
%
% This function computes a sample y from a Gaussian process with mean
% 0 and covariance K, by using the square-root method [1]. This method
% is more effective if K can be preconditioned by LL' whose spectrum
% is narrower than that of K and is far away from the origin.
%
% INPUTS
%   Kmult:  Function handle to perform K*x for a vector x.
%   Lmult:  Function handle to perform L*x. Can be empty.
%   LLTinv: Function handle to perform inv(LL')*x. Can be empty.
%   n:      Dimension of K.
%   degree: Degree of polynomial in approximating the sqrt function.
%   tol:    Tolerance in approximating the sqrt function.
%
% OUTPUTS
%   y:      Sample vector.
%
% [1] Jie Chen, Mihai Anitescu and Yousef Saad. Computing $f(A)b$ via
%     Least Squares Polynomial Approximations. SIAM Journal on
%     Scientific Computing, 33(1):195--222, 2011.
%

%Author: J. Chen
%$Date: 2011/11/23 11:49:45$

% determine the two ends of the spectrum of inv(L)*K*inv(L').
[eig_lo, n_lo, eig_hi, n_hi] = eigk(Kmult, LLTinv, n, 200, 1, 1, tol);
if n_lo == 1 & n_hi == 1
  success = 1;
else
  success = 0;
end

% polynomial interpolation
f = @sqrt;
t = get_intervals_geometric(eig_lo, eig_hi, success);
tt = interval(eig_lo, eig_hi);
[alpha, beta, gamma, res, iter] = ls_poly(f, t, tt, degree, tol);
fprintf(1, '     interp interval: [%g, %g]\n', t(1), t(end));

% generate sample y =       L * sqrt(inv(L) * K * inv(L')) * x
%                   \approx L * poly(inv(L) * K * inv(L')) * x
%                   =       poly(K * inv(LL')) * L * x
if ~isempty(Lmult) & ~isempty(LLTinv)
  y = three_term_recur(@(x)(Kmult(LLTinv(x))), Lmult(randn(n,1)), ...
                       alpha, beta, gamma);
else
  y = three_term_recur(Kmult, randn(n,1), alpha, beta, gamma);
end
