function [x, iter, res, res2, res3] = pcg(A, b, M, x0, maxit, tol, sol)
%function [x, iter, res, res2, res3] = pcg(A, b, M, x0, maxit, tol, sol)
%
% PCG: Preconditioned Conjugate Gradient
%
%                            M*Ax = M*b
%
% INPUTS
%   A:     Matrix/Operator A
%   b:     Right-hand side.
%   M:     Preconditioner M. Can be empty.
%   x0:    Initial guess.
%   maxit: Maximum number of iterations.
%   tol:   Tolerance of residual norm.
%   sol:   (Optional) Actual solution.
%
% OUTPUTS
%   x:     Computed solution.
%   iter:  Actual number of iterations.
%   res:   Residual norm |b-Ax|/|b|, all iterations.
%   res2:  (If sol available) Error norm |x-sol|_A / |b|, all iterations.
%   res3:  (If sol available) Error norm |x-sol| / |sol|, all iterations.
%

%Author: J. Chen
%$Date: 2011/11/23 11:51:27$

if isnumeric(A)
  A = @(x)(A*x);
end
if ~isempty(M) & isnumeric(M)
  M = @(x)(M*x);
end

% scale the linear system
x = x0;
normb = norm(b);
b = b/normb;
x = x/normb;
if exist('sol', 'var')
  sol = sol/normb;
  normsol = norm(sol);
end

% allocate space to hold residuals
res = zeros(maxit,1);
res2 = [];
res3 = [];
if exist('sol', 'var')
  res2 = zeros(maxit,1);
  res3 = zeros(maxit,1);
end

% initialize iterates
r = b - A(x);
if ~isempty(M)
  z = M(r);
else
  z = r;
end
p = z;
rz = r'*z;

% the first residual
res(1) = norm(r);
if exist('sol', 'var')
  e = x - sol;
  res2(1) = sqrt(abs(e'*A(e)));
  res3(1) = norm(e)/normsol;
end

% loop
iter = 1;
while iter < maxit & res(iter) > tol

  q = A(p);
  alpha = rz/(p'*q);
  x = x + alpha*p;
  r = r - alpha*q;

  rz_old = rz;
  if ~isempty(M)
    z = M(r);
  else
    z = r;
  end
  rz = r'*z;
  
  beta = rz/rz_old;
  p = z + beta*p;
  
  % compute residual
  iter = iter + 1;
  res(iter) = norm(r);
  if exist('sol', 'var')
    e = x - sol;
    res2(iter) = sqrt(abs(e'*A(e)));
    res3(iter) = norm(e)/normsol;
  end

end

% scale back the computed solution
x = x*normb;

% output simple message if fail
res = res(1:iter);
if exist('sol', 'var')
  res2 = res2(1:iter,:);
  res3 = res3(1:iter,:);
end
if iter == maxit
  fprintf(1, '  PCG: maxit reached. res = %g\n', res(end));
end
