function [Gamma, iter] = score_func_saa_covar(K, dK, M, ...
                                              y, u, m, maxit, tol, jacob)
%function [Gamma, iter] = score_func_saa_covar(K, dK, M, ...
%                                              y, u, m, maxit, tol, jacob)
%
% This function returns the covariance matrix in the convergence
% result of the SAA of score equation.
%
% INPUTS
%   K, dK:         Covariance matrix and derivatives.
%   M:             Preconditioner. Can be empty.
%   y, u:          Appear in the SAA of the score function.
%   m, maxit, tol: Parameters for the linear solver.
%   jacob:         Jacobian of the score equation.
%
% OUTPUTS
%   Gamma:         Covariance matrix in convergence result.
%   iter:          Numbers of iterations in the linear solver.
%

%Author: J. Chen
%$Date: 2011/12/19 17:45:22$

d = length(dK);

if isnumeric(K)
  K = @(x)(K*x);
end

if isnumeric(dK{1})
  for i = 1:d
    dK{i} = @(x)(dK{i}*x);
  end
end

if ~isempty(M) & isnumeric(M)
  M = @(x)(M*x);
end

N = size(u,2);

% solve K\u and K\y
B = [u y];
X0 = zeros(size(B));
n = size(B,1);
uu = randn(n,1);

if m == 0
  [Kuy, iter] = bpcg(K, B, M, X0, maxit, tol);
else
  W = lan(K, M, n, uu, m, m, m, tol);
  [Kuy, iter] = dbpcg(K, B, M, W, X0, maxit, tol);
end

% compute dK*u and dK*(K\y)
for j = 1:d
  dKu{j} = dK{j}(u);
  dKKy{j} = dK{j}(Kuy(:,N+1));
end

% assemble results
fixed = zeros(d,1);
varied = zeros(d,1);
V = zeros(d);

for j = 1:d
  fixed(j) = - Kuy(:,N+1)' * dKKy{j};
end

for i = 1:N
  for j = 1:d
    varied(j) = fixed(j) + Kuy(:,i)' * dKu{j}(:,i);
  end
  V = V + varied*varied';
end
V = V/N;

Gamma = inv(jacob)' * V * inv(jacob)/N;
Gamma = (Gamma+Gamma')/2;